# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring,redefined-outer-name


from collections import OrderedDict
import pytest
import tensorflow as tf

import rddlgym

from tfplan.planners import DEFAULT_CONFIG, StraightLinePlanner
from tfplan.planners.stochastic import utils


BATCH_SIZE = 16
HORIZON = 20
EPOCHS = 5


@pytest.fixture(scope="module", params=["Navigation-v2"])
def planner(request):
    rddl = request.param
    config = {
        **DEFAULT_CONFIG,
        **{"batch_size": BATCH_SIZE, "horizon": HORIZON, "epochs": EPOCHS},
        "verbose": False,
    }
    planner_ = StraightLinePlanner(rddl, config)
    planner_.build()
    return planner_


def test_build_policy_ops(planner):
    policy = planner.policy
    compiler = planner.compiler
    assert not policy.parallel_plans
    assert policy.horizon == compiler.rddl.instance.horizon
    assert hasattr(policy, "_policy_variables")


def test_build_initial_state_ops(planner):
    initial_state = planner.initial_state
    compiler = planner.compiler
    batch_size = compiler.batch_size
    assert isinstance(initial_state, tuple)
    assert len(initial_state) == len(compiler.initial_state_fluents)
    for tensor, fluent in zip(initial_state, compiler.initial_state_fluents):
        assert tensor.dtype == fluent[1].dtype
        assert tensor.shape == (batch_size, *fluent[1].shape.fluent_shape)


def test_build_sequence_length_ops(planner):
    assert planner.steps_to_go is not None
    assert planner.steps_to_go.dtype == tf.int32
    assert planner.steps_to_go.shape == ()
    assert planner.sequence_length is not None
    assert planner.sequence_length.dtype == tf.int32
    assert planner.sequence_length.shape == (planner.compiler.batch_size,)


def test_build_trajectory_ops(planner):
    trajectory = planner.trajectory
    actions = trajectory.actions

    batch_size = planner.compiler.batch_size
    horizon = planner.compiler.rddl.instance.horizon
    action_fluents = planner.compiler.default_action_fluents

    for action, action_fluent in zip(actions, action_fluents):
        size = action_fluent[1].shape.as_list()
        assert action.shape.as_list() == [batch_size, horizon, *size]


def test_call(planner):
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)
    state, timestep = env.reset()
    action = planner(state, timestep)
    assert isinstance(action, OrderedDict)


def test_get_batch_initial_state(planner):
    # pylint: disable=protected-access
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)

    with planner.compiler.graph.as_default():
        state = env.observation_space.sample()
        batch_state = planner._get_batch_initial_state(state)
        assert len(state) == len(batch_state)

        for fluent, batch_fluent in zip(state.values(), batch_state):
            assert fluent.dtype == batch_fluent.dtype
            assert fluent.shape == batch_fluent.shape[1:]
            assert batch_fluent.shape[0] == planner.compiler.batch_size


def test_get_action(planner):
    # pylint: disable=protected-access
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)

    with tf.Session(graph=planner.compiler.graph) as sess:
        sess.run(tf.global_variables_initializer())
        state = env.observation_space.sample()
        batch_state = planner._get_batch_initial_state(state)
        samples = utils.evaluate_noise_samples_as_inputs(
            sess, planner.simulator.samples
        )
        feed_dict = {
            planner.initial_state: batch_state,
            planner.simulator.noise: samples,
            planner.steps_to_go: HORIZON,
        }
        actions_ = planner._get_action(planner.trajectory.actions, feed_dict)
        action_fluents = planner.compiler.default_action_fluents
        assert isinstance(actions_, OrderedDict)
        assert len(actions_) == len(action_fluents)
        for action_, action_fluent in zip(actions_.values(), action_fluents):
            assert tf.dtypes.as_dtype(action_.dtype) == action_fluent[1].dtype
            assert list(action_.shape) == list(action_fluent[1].shape.fluent_shape)


def test_runner(planner):
    # pylint: disable=protected-access
    rddl = planner.rddl
    env = rddlgym.make(rddl, mode=rddlgym.GYM)
    env._horizon = 3
    runner = rddlgym.Runner(env, planner)
    trajectory = runner.run()
    assert len(trajectory) == env._horizon



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


import tensorflow as tf


OPTIMIZERS = {
    "Adadelta": tf.compat.v1.train.AdadeltaOptimizer,
    "Adagrad": tf.compat.v1.train.AdagradOptimizer,
    "Adam": tf.compat.v1.train.AdamOptimizer,
    "GradientDescent": tf.compat.v1.train.GradientDescentOptimizer,
    "ProximalGradientDescent": tf.compat.v1.train.ProximalGradientDescentOptimizer,
    "ProximalAdagrad": tf.compat.v1.train.ProximalAdagradOptimizer,
    "RMSProp": tf.compat.v1.train.RMSPropOptimizer,
}


DEFAULT_CONFIG = {"optimizer": "RMSProp", "learning_rate": 0.001}


class ActionOptimizer:
    """ActionOptimizer wraps a TensorFlow optimizer.

    Args:
        config (Dict[str, Any]): The optimizer config dict.
    """

    def __init__(self, config):
        self.config = config

        self.optimizer = None

    def build(self):
        """Builds the underlying optimizer."""
        tf_optimizer = OPTIMIZERS[self.config["optimizer"]]
        learning_rate = self.config["learning_rate"]
        self.optimizer = tf_optimizer(learning_rate)

    def compute_gradients(self, loss):
        return self.optimizer.compute_gradients(loss)

    def apply_gradients(self, grads_and_vars):
        return self.optimizer.apply_gradients(grads_and_vars)

    def minimize(self, loss):
        """Returns the train op corresponding to the loss minimization."""
        return self.optimizer.minimize(loss)



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


from collections import OrderedDict
import os

import tensorflow as tf
from tqdm import trange

from rddl2tf.compilers import DefaultCompiler

from tfplan.planners.planner import Planner
from tfplan.planners.deterministic.simulation import Simulator
from tfplan.train.policy import OpenLoopPolicy
from tfplan.train.optimizer import ActionOptimizer


class Tensorplan(Planner):
    """Tensorplan class implements the Planner interface
    for the offline gradient-based planner (i.e., tensorplan).

    Args:
        model (pyrddl.rddl.RDDL): A RDDL model.
        config (Dict[str, Any]): The planner config dict.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, rddl, config):
        super(Tensorplan, self).__init__(rddl, DefaultCompiler, config)

        self.policy = None
        self.initial_state = None

        self.simulator = None
        self.trajectory = None
        self.final_state = None
        self.total_reward = None

        self.avg_total_reward = None
        self.loss = None

        self.optimizer = None
        self.train_op = None

        self.best_plan_idx = None
        self.best_plan = None

        self._plan = None

        self.writer = None
        self.summaries = None

    @property
    def logdir(self):
        return self.config.get("logdir") or f"/tmp/tfplan/tensorplan/{self.rddl}"

    def build(self):
        """Builds planner ops."""
        with self.graph.as_default():
            self._build_policy_ops()
            self._build_initial_state_ops()
            self._build_trajectory_ops()
            self._build_loss_ops()
            self._build_optimization_ops()
            self._build_solution_ops()
            self._build_summary_ops()
            self._build_init_ops()

    def _build_init_ops(self):
        self.init_op = tf.global_variables_initializer()

    def _build_policy_ops(self):
        horizon = self.config["horizon"]
        self.policy = OpenLoopPolicy(self.compiler, horizon, parallel_plans=True)
        self.policy.build("tensorplan")

    def _build_initial_state_ops(self):
        self.initial_state = self.compiler.initial_state()

    def _build_trajectory_ops(self):
        self.simulator = Simulator(self.compiler, self.policy)
        self.simulator.build()
        self.trajectory, self.final_state, self.total_reward = self.simulator.trajectory(
            self.initial_state
        )

    def _build_loss_ops(self):
        with tf.name_scope("loss"):
            self.avg_total_reward = tf.reduce_mean(self.total_reward)
            self.loss = tf.square(self.avg_total_reward)

    def _build_optimization_ops(self):
        self.optimizer = ActionOptimizer(self.config["optimization"])
        self.optimizer.build()
        self.train_op = self.optimizer.minimize(self.loss)

    def _build_solution_ops(self):
        self.best_plan_idx = tf.argmax(self.total_reward, axis=0)
        self.best_plan = tuple(
            action[self.best_plan_idx] for action in self.trajectory.actions
        )

    def _build_summary_ops(self):
        tf.compat.v1.summary.histogram("total_reward", self.total_reward)
        tf.compat.v1.summary.scalar("avg_total_reward", self.avg_total_reward)
        tf.compat.v1.summary.scalar("loss", self.loss)
        self.summaries = tf.compat.v1.summary.merge_all()

    def run(self):
        """Run the planner for the given number of epochs.

        Returns:
            plan (Sequence(np.ndarray): The best solution plan.
        """
        self.writer = tf.compat.v1.summary.FileWriter(self.logdir, self.graph)

        self._sess.run(self.init_op)

        run_id = self.config.get("run_id", 0)
        pid = os.getpid()
        position = run_id % self.config.get("num_workers", 1)
        epochs = self.config["epochs"]
        desc = f"(pid={pid}) Run #{run_id:<3d}"

        with trange(
            epochs, desc=desc, unit="epoch", position=position, leave=False
        ) as t:

            for step in t:
                _, loss_, avg_total_reward_, summary_ = self._sess.run(
                    [self.train_op, self.loss, self.avg_total_reward, self.summaries]
                )

                self.writer.add_summary(summary_, step)

                t.set_postfix(
                    loss=f"{loss_:10.4f}", avg_total_reward=f"{avg_total_reward_:10.4f}"
                )

        self.writer.close()

        plan_ = self._sess.run(self.best_plan)
        return plan_

    def __call__(self, state, timestep):
        """Returns the action for the given `timestep`."""
        # find plan
        if self._plan is None:
            self._plan = self.run()

        # select action for given timestep
        action_fluent_ordering = self.compiler.rddl.domain.action_fluent_ordering
        action = OrderedDict(
            {
                name: action[timestep]
                for name, action in zip(action_fluent_ordering, self._plan)
            }
        )
        return action



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.


# pylint: disable=missing-docstring,redefined-outer-name


import pytest
import numpy as np
import tensorflow as tf

import rddlgym
from rddl2tf.compilers import ReparameterizationCompiler

from tfplan.planners.stochastic import utils
from tfplan.planners.stochastic.simulation import SimulationCell, Simulator
from tfplan.train.policy import OpenLoopPolicy


BATCH_SIZE = 32
HORIZON = 20


@pytest.fixture(scope="module", params=["Navigation-v3"])
def cell(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)

    compiler = ReparameterizationCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()

    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=True)
    policy.build("planning")

    with compiler.graph.as_default():
        reparameterization_map = compiler.get_cpfs_reparameterization()
        cell_samples = utils.get_noise_samples(
            reparameterization_map, BATCH_SIZE, horizon=1
        )
        cell_noise, encoding = utils.encode_noise_samples_as_inputs(cell_samples)

    cell = SimulationCell(compiler, policy, config={"encoding": encoding})
    cell.cell_noise = cell_noise
    yield cell


def test_state_size(cell):
    rddl = cell.compiler.rddl
    state_size = rddl.state_size
    cell_state_size = cell.state_size
    assert len(cell_state_size) == len(state_size)
    for size, cell_size in zip(state_size, cell_state_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_action_size(cell):
    rddl = cell.compiler.rddl
    action_size = rddl.action_size
    cell_action_size = cell.action_size
    assert len(cell_action_size) == len(action_size)
    for size, cell_size in zip(action_size, cell_action_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_interm_size(cell):
    rddl = cell.compiler.rddl
    interm_size = rddl.interm_size
    cell_interm_size = cell.interm_size
    assert len(cell_interm_size) == len(interm_size)
    for size, cell_size in zip(interm_size, cell_interm_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_output_size(cell):
    output_size = cell.output_size
    assert output_size == (cell.state_size, cell.action_size, cell.interm_size, 1)


def test_call(cell):
    state = cell.compiler.initial_state()
    action = cell.compiler.default_action()

    with cell.compiler.graph.as_default():
        timesteps = tf.zeros((BATCH_SIZE, 1), dtype=tf.float32)
        inputs = tf.concat([timesteps, cell.cell_noise[:, 0, ...]], axis=1)
        output, _ = cell(inputs, state)

    assert len(output) == 4
    assert output[0] is not None
    assert output[1] is not None
    assert output[2] is not None
    assert output[3] is not None

    assert len(output[0]) == len(state)
    for output_state_tensor, state_tensor in zip(output[0], state):
        assert isinstance(output_state_tensor[0], tf.Tensor)
        assert output_state_tensor[0].dtype == state_tensor.dtype
        assert output_state_tensor[0].shape == state_tensor.shape

    assert len(output[1]) == len(action)
    for output_action_tensor, action_tensor in zip(output[1], action):
        assert isinstance(output_action_tensor[0], tf.Tensor)
        assert output_action_tensor[0].dtype == action_tensor.dtype
        assert output_action_tensor[0].shape == action_tensor.shape

    interm_size = cell.compiler.rddl.interm_size
    assert len(output[2]) == len(interm_size)
    for output_interm_tensor, interm_fluent_size in zip(output[2], interm_size):
        assert isinstance(output_interm_tensor[0], tf.Tensor)
        assert output_interm_tensor[0].shape == (BATCH_SIZE, *interm_fluent_size)

    assert isinstance(output[3], tf.Tensor)
    assert output[3].dtype == tf.float32
    assert output[3].shape == (BATCH_SIZE, 1)


@pytest.fixture(scope="module", params=["Navigation-v3"])
def simulator(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)

    compiler = ReparameterizationCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()

    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=True)
    policy.build("planning")

    simulator = Simulator(compiler, policy, config=None)
    simulator.build()
    yield simulator


def test_build(simulator):
    reparameterization_map = simulator.cell
    assert reparameterization_map is not None

    samples = simulator.samples
    assert samples is not None

    noise = simulator.noise
    assert noise is not None

    cell = simulator.cell
    assert cell is not None
    assert isinstance(cell, SimulationCell)


def test_trajectory(simulator):
    state = simulator.cell.compiler.initial_state()
    action = simulator.cell.compiler.default_action()

    trajectory, final_state, total_reward = simulator.trajectory(state)

    assert len(trajectory) == 4
    assert trajectory.states is not None
    assert trajectory.actions is not None
    assert trajectory.interms is not None
    assert trajectory.rewards is not None

    assert len(trajectory.states) == len(state)
    for output_state_tensor, state_tensor in zip(trajectory.states, state):
        assert isinstance(output_state_tensor, tf.Tensor)
        assert output_state_tensor.dtype == state_tensor.dtype
        assert output_state_tensor.shape == (
            BATCH_SIZE,
            HORIZON,
            *state_tensor.shape[1:],
        )

    assert len(trajectory.actions) == len(action)
    for output_action_tensor, action_tensor in zip(trajectory.actions, action):
        assert isinstance(output_action_tensor, tf.Tensor)
        assert output_action_tensor.dtype == action_tensor.dtype
        assert output_action_tensor.shape == (
            BATCH_SIZE,
            HORIZON,
            *action_tensor.shape[1:],
        )

    interm_size = simulator.cell.compiler.rddl.interm_size
    assert len(trajectory.interms) == len(interm_size)
    for output_interm_tensor, interm_fluent_size in zip(
        trajectory.interms, interm_size
    ):
        assert isinstance(output_interm_tensor, tf.Tensor)
        assert output_interm_tensor.shape == (BATCH_SIZE, HORIZON, *interm_fluent_size)

    assert isinstance(trajectory.rewards, tf.Tensor)
    assert trajectory.rewards.dtype == tf.float32
    assert trajectory.rewards.shape == (BATCH_SIZE, HORIZON, 1)

    assert len(final_state) == len(state)
    for output_state_tensor, state_tensor in zip(final_state, state):
        assert isinstance(output_state_tensor, tf.Tensor)
        assert output_state_tensor.dtype == state_tensor.dtype
        assert output_state_tensor.shape == (BATCH_SIZE, *state_tensor.shape[1:])

    assert isinstance(total_reward, tf.Tensor)
    assert total_reward.dtype == tf.float32
    assert total_reward.shape == (BATCH_SIZE,)


def test_run(simulator):
    state = simulator.cell.compiler.initial_state()
    trajectory, _, _ = simulator.trajectory(state)
    trajectory_ = simulator.run(trajectory)
    assert trajectory_ is not None


def test_timesteps():
    timesteps = Simulator.timesteps(BATCH_SIZE, HORIZON)
    assert isinstance(timesteps, tf.Tensor)
    assert timesteps.dtype == tf.float32
    assert timesteps.shape == (BATCH_SIZE, HORIZON, 1)
    with tf.Session() as sess:
        timesteps_ = sess.run(timesteps)
        for batch in range(BATCH_SIZE):
            assert np.allclose(timesteps_[batch], timesteps_[0])
            steps = timesteps_[batch]
            for time in range(1, HORIZON):
                assert steps[time] == steps[time - 1] + 1



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.


# pylint: disable=missing-docstring


from collections import namedtuple
import tensorflow as tf

from tfplan.planners.deterministic import utils


OutputTuple = namedtuple("OutputTuple", "state action interm reward")
Trajectory = namedtuple("Trajectory", "states actions interms rewards")


class SimulationCell(tf.compat.v1.nn.rnn_cell.RNNCell):
    """SimulationCell class implements an RNN cell that simulates the
    next state and reward for the MDP transition given by the RDDL model.

    Args:
        compiler (rddl2tf.compilers.DefaulCompiler): The RDDL2TF compiler.
        policy (tfplan.train.OpenLoopPolicy): The state-independent policy (e.g., a plan).
    """

    def __init__(self, compiler, policy):
        self.compiler = compiler
        self.policy = policy

    @property
    def state_size(self):
        """Returns the MDP state size."""
        return utils.cell_size(self.compiler.rddl.state_size)

    @property
    def action_size(self):
        """Returns the MDP action size."""
        return utils.cell_size(self.compiler.rddl.action_size)

    @property
    def interm_size(self):
        """Returns the MDP intermediate state size."""
        return utils.cell_size(self.compiler.rddl.interm_size)

    @property
    def output_size(self):
        """Returns the simulation cell output size."""
        return (self.state_size, self.action_size, self.interm_size, 1)

    def __call__(self, inputs, state, scope=None):
        """Returns the cell's output tuple and next state tensors.

        Output tuple packs together the next state, action, interms,
        and reward tensors in order.

        Args:
            inputs (tf.Tensor): The encoded (timestep, noise) input tensor.
            state (Sequence[tf.Tensor]): The current state tensors.
            scope (Optional[str]): The cell name scope.

        Returns:
            (CellOutput, CellState): A pair with the cell's output tuple and next state.
        """

        # timestep
        timestep = inputs

        # action
        action = self.policy(state, timestep)

        # next state
        interm, next_state = self.compiler.cpfs(state, action)

        # reward
        reward = self.compiler.reward(state, action, next_state)

        # outputs
        next_state = utils.to_tensor(next_state)
        interm = utils.to_tensor(interm)
        output = OutputTuple(next_state, action, interm, reward)

        return (output, next_state)


class Simulator:
    """Simulator class implements an RNN-based trajctory simulator
    for the RDDL model.

    Args:
        compiler (rddl2tf.compilers.DefaulCompiler): The RDDL2TF compiler.
        policy (tfplan.train.OpenLoopPolicy): The state-independent policy (e.g., a plan).
    """

    def __init__(self, compiler, policy):
        self.compiler = compiler
        self.policy = policy

        self.cell = None
        self.inputs = None

    @property
    def graph(self):
        """Returns the compiler's graph."""
        return self.compiler.graph

    @property
    def batch_size(self):
        """Returns the policy's batch size."""
        return self.policy.batch_size

    @property
    def horizon(self):
        """Returns the policy's batch size."""
        return self.policy.horizon

    def build(self):
        """Builds the recurrent cell ops by embedding the
        policy in the transition sampling cell.
        """
        self.cell = SimulationCell(self.compiler, self.policy)

    def trajectory(self, initial_state):
        """Returns the state-action-reward trajectory induced by
        the given `initial_state` and policy.

        Args:
            initial_state (Sequence[tf.Tensor]): The trajectory's initial state.

        Returns:
            trajectory (Trajectory): The collection of states-actions-interms-rewards trajectory.
            final_state (Sequence[tf.Tensor]): The trajectory's final state.
            total_reward (tf.Tensor(shape=(batch_size,))): The trajectory's total reward.
        """
        with self.graph.as_default():

            with tf.name_scope("inputs"):
                self.inputs = self.timesteps(self.batch_size, self.horizon)

            with tf.name_scope("trajectory"):
                outputs, final_state = tf.nn.dynamic_rnn(
                    self.cell,
                    self.inputs,
                    initial_state=initial_state,
                    dtype=tf.float32,
                )

            with tf.name_scope("total_reward"):
                total_reward = tf.reduce_sum(tf.squeeze(outputs[3]), axis=1)

        states = tuple(fluent[0] for fluent in outputs[0])
        actions = tuple(fluent[0] for fluent in outputs[1])
        interms = tuple(fluent[0] for fluent in outputs[2])
        rewards = outputs[3]
        trajectory = Trajectory(states, actions, interms, rewards)

        return trajectory, final_state, total_reward

    def run(self, trajectory):
        """Evaluates the given `trajectory`."""
        with tf.Session(graph=self.graph) as sess:
            sess.run(tf.global_variables_initializer())
            return sess.run(trajectory)

    @classmethod
    def timesteps(cls, batch_size, horizon):
        """Returns the batch-sized decreasing-horizon timesteps tensor."""
        with tf.name_scope("timesteps"):
            start, limit, delta = horizon - 1, -1, -1
            timesteps_range = tf.range(start, limit, delta, dtype=tf.float32)
            timesteps_range = tf.expand_dims(timesteps_range, -1)
            batch_timesteps = tf.stack([timesteps_range] * batch_size)
            return batch_timesteps



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.


"""Collection of RNN-based simulation utility functions."""

# pylint: disable=missing-docstring


def cell_size(sizes):
    return tuple(sz if sz != () else (1,) for sz in sizes)


def to_tensor(fluents):
    return tuple(f.tensor for f in fluents)



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


import abc

import tensorflow as tf

import rddlgym


DEFAULT_CONFIG = {
    "batch_size": 128,
    "epochs": 200,
    "optimization": {"optimizer": "RMSProp", "learning_rate": 0.001},
}


class Planner(metaclass=abc.ABCMeta):
    """Planner abstract base class.

    Args:
        rddl: (pyrddl.rddl.RDDL): A RDDL
        compiler_cls (rddl2tf.Compiler): The RDDL-to-TensorFlow compiler class.
        config (Dict[str, Any]): The planner config dict.
    """

    def __init__(self, rddl, compiler_cls, config):
        self.rddl = rddl
        self.model = rddlgym.make(rddl, mode=rddlgym.AST)
        self.compiler = compiler_cls(self.model, batch_size=config["batch_size"])
        self.config = config

        self.compiler.init()

        config = tf.ConfigProto(
            inter_op_parallelism_threads=1,
            intra_op_parallelism_threads=1,
            log_device_placement=False,
        )
        self._sess = tf.Session(graph=self.graph, config=config)

    @property
    def graph(self):
        """Returns the compiler's graph."""
        return self.compiler.graph

    @property
    def batch_size(self):
        """Returns the compiler's batch size."""
        return self.compiler.batch_size

    @abc.abstractmethod
    def build(self):
        """Builds the planner."""
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, state, timestep):
        raise NotImplementedError

    def close(self):
        self._sess.close()



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring

import abc
import collections
import os

import numpy as np
import tensorflow as tf
from tqdm import trange

from rddl2tf.compilers import ReparameterizationCompiler

from tfplan.planners.planner import Planner
from tfplan.train.optimizer import ActionOptimizer


class StochasticPlanner(Planner):
    """StochasticPlanner abstract class implements basic methods for
    online stochastic gradient-based planners.

    Args:
        rddl (str): A RDDL domain/instance filepath or rddlgym id.
        compiler_cls (rddl2tf.Compiler): The RDDL-to-TensorFlow compiler class.
        config (Dict[str, Any]): The planner config dict.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, rddl, compiler_cls, config):
        super().__init__(rddl, ReparameterizationCompiler, config)

        self.initial_state = None

        self.steps_to_go = None
        self.sequence_length = None

        self.optimizer = None
        self.grads_and_vars = None

        self.avg_total_reward = None
        self.loss = None

        self.init_op = None
        self.train_op = None

        self.summaries = None

    def build(self,):
        with self.graph.as_default():
            self._build_policy_ops()
            self._build_initial_state_ops()
            self._build_sequence_length_ops()
            self._build_trajectory_ops()
            self._build_loss_ops()
            self._build_optimization_ops()
            self._build_summary_ops()
            self._build_init_ops()

    @abc.abstractmethod
    def __call__(self, state, timestep):
        raise NotImplementedError

    @abc.abstractmethod
    def _build_policy_ops(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _build_trajectory_ops(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _build_loss_ops(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _build_summary_ops(self):
        raise NotImplementedError

    def _build_init_ops(self):
        self.init_op = tf.global_variables_initializer()

    def _build_initial_state_ops(self):
        with tf.name_scope("initial_state"):
            self.initial_state = tuple(
                tf.placeholder(t.dtype, t.shape) for t in self.compiler.initial_state()
            )

    def _build_sequence_length_ops(self):
        with tf.name_scope("sequence_length"):
            self.steps_to_go = tf.placeholder(tf.int32, shape=())
            self.sequence_length = tf.tile(
                tf.reshape(self.steps_to_go, [1]), [self.batch_size]
            )

    def _build_optimization_ops(self):
        with tf.name_scope("optimization"):
            self.optimizer = ActionOptimizer(self.config["optimization"])
            self.optimizer.build()
            self.grads_and_vars = self.optimizer.compute_gradients(self.loss)
            self.train_op = self.optimizer.apply_gradients(self.grads_and_vars)

    def _get_batch_initial_state(self, state):
        batch_size = self.compiler.batch_size
        return tuple(
            map(
                lambda fluent: np.tile(
                    fluent, (batch_size, *([1] * len(fluent.shape)))
                ),
                state.values(),
            )
        )

    def _get_action(self, actions, feed_dict):
        action_fluent_ordering = self.compiler.rddl.domain.action_fluent_ordering
        actions = self._sess.run(actions, feed_dict=feed_dict)
        action = collections.OrderedDict(
            {
                name: fluent[0][0]
                for name, fluent in zip(action_fluent_ordering, actions)
            }
        )
        return action

    def run(self, timestep, feed_dict):
        self._sess.run(self.init_op)

        if self.summaries:
            logdir = os.path.join(self.config.get("logdir"), f"timestep={timestep}")
            writer = tf.compat.v1.summary.FileWriter(logdir)

        run_id = self.config.get("run_id", 0)
        pid = os.getpid()
        position = run_id % self.config.get("num_workers", 1)
        epochs = self.config["epochs"]
        desc = f"(pid={pid}) Run #{run_id:<3d} / step={timestep:<3d}"

        with trange(
            epochs, desc=desc, unit="epoch", position=position, leave=False
        ) as t:

            for step in t:
                _, loss_, avg_total_reward_ = self._sess.run(
                    [self.train_op, self.loss, self.avg_total_reward],
                    feed_dict=feed_dict,
                )

                if self.summaries:
                    summary_ = self._sess.run(self.summaries, feed_dict=feed_dict)
                    writer.add_summary(summary_, step)

                t.set_postfix(
                    loss=f"{loss_:10.4f}", avg_total_reward=f"{avg_total_reward_:10.4f}"
                )

        if self.summaries:
            writer.close()



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring,redefined-outer-name,protected-access


import numpy as np
import pytest
import tensorflow as tf

import rddlgym
from rddl2tf.compilers import DefaultCompiler

from tfplan.train.policy import OpenLoopPolicy


HORIZON = 20
BATCH_SIZE = 32


@pytest.fixture(scope="module", params=["Navigation-v1"])
def compiler(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)
    compiler = DefaultCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()
    return compiler


@pytest.fixture(scope="module")
def parallel_plans(compiler):
    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=True)
    policy.build("parallel_plans")
    return policy


@pytest.fixture(scope="module")
def non_parallel_plans(compiler):
    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=False)
    policy.build("non_parallel_plans")
    return policy


def test_build_policy_variables(parallel_plans, non_parallel_plans):
    _test_build_policy_variables(parallel_plans, batch_size=BATCH_SIZE)
    _test_build_policy_variables(non_parallel_plans, batch_size=1)


def _test_build_policy_variables(policy, batch_size):
    compiler = policy._compiler
    action_size = compiler.rddl.action_size
    policy_variables = policy._policy_variables

    assert isinstance(policy_variables, tuple)
    assert len(policy_variables) == len(action_size)

    for policy_var, size in zip(policy_variables, action_size):
        assert isinstance(policy_var, tf.Variable)
        assert policy_var.shape == (batch_size, HORIZON, *size)


def test_call_parallel_plans(parallel_plans):
    policy_variables = parallel_plans._policy_variables

    compiler = parallel_plans._compiler
    state = compiler.initial_state()
    with compiler.graph.as_default():
        timestep = tf.constant(0, dtype=tf.int32, shape=(BATCH_SIZE, 1))

    action = parallel_plans(state, timestep)

    assert len(action) == len(policy_variables)
    for action_tensor, policy_var in zip(action, policy_variables):
        assert isinstance(action_tensor, tf.Tensor)
        assert action_tensor.shape == (BATCH_SIZE, *policy_var.shape[2:])

    with tf.Session(graph=compiler.graph) as sess:
        sess.run(tf.global_variables_initializer())
        actions_ = sess.run(action)

        for action_ in actions_:
            for i in range(BATCH_SIZE):
                for j in range(BATCH_SIZE):
                    if i != j:
                        assert not np.allclose(action_[i], action_[j])


def test_call_non_parallel_plans(non_parallel_plans):
    policy_variables = non_parallel_plans._policy_variables

    compiler = non_parallel_plans._compiler
    state = compiler.initial_state()
    with compiler.graph.as_default():
        timestep = tf.constant(0, dtype=tf.int32, shape=(BATCH_SIZE, 1))

    action = non_parallel_plans(state, timestep)

    assert len(action) == len(policy_variables)
    for action_tensor, policy_var in zip(action, policy_variables):
        assert isinstance(action_tensor, tf.Tensor)
        assert action_tensor.shape == (BATCH_SIZE, *policy_var.shape[2:])

    with tf.Session(graph=compiler.graph) as sess:
        sess.run(tf.global_variables_initializer())
        actions_ = sess.run(action)

        for action_ in actions_:
            for i in range(BATCH_SIZE):
                for j in range(BATCH_SIZE):
                    assert np.allclose(action_[i], action_[j])



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring,redefined-outer-name


import numpy as np
import pytest

import rddlgym

from tfplan.planners import Tensorplan
from tfplan.train.policy import OpenLoopPolicy


HORIZON = 20
BATCH_SIZE = 32
EPOCHS = 10


@pytest.fixture(scope="module", params=["Navigation-v1"])
def planner(request):
    rddl = request.param
    config = {
        "batch_size": BATCH_SIZE,
        "epochs": EPOCHS,
        "horizon": HORIZON,
        "optimization": {"optimizer": "RMSProp", "learning_rate": 1e-3},
    }
    planner = Tensorplan(rddl, config)
    planner.build()
    return planner


def test_build_policy_ops(planner):
    assert planner.policy is not None
    assert isinstance(planner.policy, OpenLoopPolicy)
    assert planner.policy.parallel_plans


def test_build_initial_state_ops(planner):
    assert planner.initial_state is not None


def test_build_trajectory_ops(planner):
    assert planner.trajectory is not None
    assert planner.final_state is not None
    assert planner.total_reward is not None


def test_build_loss_ops(planner):
    assert planner.avg_total_reward is not None
    assert planner.loss is not None


def test_build_optimization_ops(planner):
    assert planner.optimizer is not None
    assert planner.train_op is not None


def test_build_solution_ops(planner):
    assert planner.best_plan_idx is not None
    assert planner.best_plan is not None


def test_run(planner):
    action = planner.compiler.default_action()

    plan_ = planner.run()
    assert plan_ is not None

    assert len(plan_) == len(action)
    for plan_action, action_tensor in zip(plan_, action):
        assert isinstance(plan_action, np.ndarray)
        assert plan_action.shape == (HORIZON, *action_tensor.shape[1:])


def test_call(planner):
    # pylint: disable=invalid-name,protected-access
    state = None
    for timestep in range(HORIZON):
        action = planner(state, timestep)

        assert len(action) == len(planner._plan)
        for (_, a1), a2 in zip(action.items(), planner._plan):
            assert np.allclose(a1, a2[timestep])


def test_runner(planner):
    # pylint: disable=protected-access
    rddl = planner.rddl
    env = rddlgym.make(rddl, mode=rddlgym.GYM)
    env.set_horizon(HORIZON)
    runner = rddlgym.Runner(env, planner)
    trajectory = runner.run()
    assert len(trajectory) == env._horizon



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.


# pylint: disable=missing-docstring,redefined-outer-name


import pytest
import numpy as np
import tensorflow as tf

import rddlgym
from rddl2tf.compilers import DefaultCompiler

from tfplan.planners.deterministic.simulation import SimulationCell, Simulator
from tfplan.train.policy import OpenLoopPolicy


HORIZON = 20
BATCH_SIZE = 32


@pytest.fixture(scope="module", params=["Navigation-v1"])
def cell(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)
    compiler = DefaultCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()
    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=True)
    policy.build("tensorplan")
    yield SimulationCell(compiler, policy)


def test_state_size(cell):
    rddl = cell.compiler.rddl
    state_size = rddl.state_size
    cell_state_size = cell.state_size
    assert len(cell_state_size) == len(state_size)
    for size, cell_size in zip(state_size, cell_state_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_action_size(cell):
    rddl = cell.compiler.rddl
    action_size = rddl.action_size
    cell_action_size = cell.action_size
    assert len(cell_action_size) == len(action_size)
    for size, cell_size in zip(action_size, cell_action_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_interm_size(cell):
    rddl = cell.compiler.rddl
    interm_size = rddl.interm_size
    cell_interm_size = cell.interm_size
    assert len(cell_interm_size) == len(interm_size)
    for size, cell_size in zip(interm_size, cell_interm_size):
        if size == tuple():
            assert cell_size == (1,)
        else:
            assert cell_size == size


def test_output_size(cell):
    output_size = cell.output_size
    assert output_size == (cell.state_size, cell.action_size, cell.interm_size, 1)


def test_call(cell):
    state = cell.compiler.initial_state()
    action = cell.compiler.default_action()

    with cell.compiler.graph.as_default():
        inputs = Simulator.timesteps(BATCH_SIZE, HORIZON)
        inputs = inputs[:, 0, ...]

    output, next_state = cell(inputs, state)

    assert len(output) == 4
    assert output.state is not None
    assert output.action is not None
    assert output.interm is not None
    assert output.reward is not None

    assert output.state is next_state

    assert len(output.state) == len(state)
    for output_state_tensor, state_tensor in zip(output.state, state):
        assert isinstance(output_state_tensor, tf.Tensor)
        assert output_state_tensor.dtype == state_tensor.dtype
        assert output_state_tensor.shape == state_tensor.shape

    assert len(output.action) == len(action)
    for output_action_tensor, action_tensor in zip(output.action, action):
        assert isinstance(output_action_tensor, tf.Tensor)
        assert output_action_tensor.dtype == action_tensor.dtype
        assert output_action_tensor.shape == action_tensor.shape

    interm_size = cell.compiler.rddl.interm_size
    assert len(output.interm) == len(interm_size)
    for output_interm_tensor, interm_fluent_size in zip(output.interm, interm_size):
        assert isinstance(output_interm_tensor, tf.Tensor)
        assert output_interm_tensor.shape == (BATCH_SIZE, *interm_fluent_size)

    assert isinstance(output.reward, tf.Tensor)
    assert output.reward.dtype == tf.float32
    assert output.reward.shape == (BATCH_SIZE, 1)


@pytest.fixture(scope="module", params=["Navigation-v1"])
def simulator(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)
    compiler = DefaultCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()
    policy = OpenLoopPolicy(compiler, HORIZON, parallel_plans=True)
    policy.build("tensorplan")
    simulator = Simulator(compiler, policy)
    simulator.build()
    yield simulator


def test_build(simulator):
    cell = simulator.cell
    assert cell is not None
    assert isinstance(cell, SimulationCell)


def test_trajectory(simulator):
    state = simulator.cell.compiler.initial_state()
    action = simulator.cell.compiler.default_action()

    trajectory, final_state, total_reward = simulator.trajectory(state)

    assert len(trajectory) == 4
    assert trajectory.states is not None
    assert trajectory.actions is not None
    assert trajectory.interms is not None
    assert trajectory.rewards is not None

    assert len(trajectory.states) == len(state)
    for output_state_tensor, state_tensor in zip(trajectory.states, state):
        assert isinstance(output_state_tensor, tf.Tensor)
        assert output_state_tensor.dtype == state_tensor.dtype
        assert output_state_tensor.shape == (
            BATCH_SIZE,
            HORIZON,
            *state_tensor.shape[1:],
        )

    assert len(trajectory.actions) == len(action)
    for output_action_tensor, action_tensor in zip(trajectory.actions, action):
        assert isinstance(output_action_tensor, tf.Tensor)
        assert output_action_tensor.dtype == action_tensor.dtype
        assert output_action_tensor.shape == (
            BATCH_SIZE,
            HORIZON,
            *action_tensor.shape[1:],
        )

    interm_size = simulator.cell.compiler.rddl.interm_size
    assert len(trajectory.interms) == len(interm_size)
    for output_interm_tensor, interm_fluent_size in zip(
        trajectory.interms, interm_size
    ):
        assert isinstance(output_interm_tensor, tf.Tensor)
        assert output_interm_tensor.shape == (BATCH_SIZE, HORIZON, *interm_fluent_size)

    assert isinstance(trajectory.rewards, tf.Tensor)
    assert trajectory.rewards.dtype == tf.float32
    assert trajectory.rewards.shape == (BATCH_SIZE, HORIZON, 1)

    assert len(final_state) == len(state)
    for output_state_tensor, state_tensor in zip(final_state, state):
        assert isinstance(output_state_tensor, tf.Tensor)
        assert output_state_tensor.dtype == state_tensor.dtype
        assert output_state_tensor.shape == (BATCH_SIZE, *state_tensor.shape[1:])

    assert isinstance(total_reward, tf.Tensor)
    assert total_reward.dtype == tf.float32
    assert total_reward.shape == (BATCH_SIZE,)


def test_run(simulator):
    state = simulator.cell.compiler.initial_state()
    trajectory, _, _ = simulator.trajectory(state)
    trajectory_ = simulator.run(trajectory)
    assert trajectory_ is not None


def test_timesteps():
    timesteps = Simulator.timesteps(BATCH_SIZE, HORIZON)
    assert isinstance(timesteps, tf.Tensor)
    assert timesteps.dtype == tf.float32
    assert timesteps.shape == (BATCH_SIZE, HORIZON, 1)
    with tf.Session() as sess:
        timesteps_ = sess.run(timesteps)
        for batch in range(BATCH_SIZE):
            assert np.allclose(timesteps_[batch], timesteps_[0])
            steps = timesteps_[batch]
            for time in range(1, HORIZON):
                assert steps[time] == steps[time - 1] - 1



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring,redefined-outer-name,invalid-name


from collections import namedtuple
import pytest
import numpy as np
import tensorflow as tf

import rddlgym
from rddl2tf.compilers.modes.reparameterization import ReparameterizationCompiler

from tfplan.planners.stochastic import utils


BATCH_SIZE = 32
HORIZON = 20


Reparameterization = namedtuple(
    "Reparameterization", "compiler mapping samples inputs encoding decoded_samples"
)


@pytest.fixture(scope="module", params=["Navigation-v3", "Reservoir-8", "HVAC-3"])
def reparameterization(request):
    rddl = request.param
    model = rddlgym.make(rddl, mode=rddlgym.AST)
    compiler = ReparameterizationCompiler(model, batch_size=BATCH_SIZE)
    compiler.init()

    with compiler.graph.as_default():
        mapping = compiler.get_cpfs_reparameterization()
        samples = utils.get_noise_samples(mapping, BATCH_SIZE, HORIZON)

        inputs, encoding = utils.encode_noise_samples_as_inputs(samples)

        decoded_samples = utils.decode_inputs_as_noise_samples(
            inputs[:, 0, ...], encoding
        )

    return Reparameterization(
        compiler, mapping, samples, inputs, encoding, decoded_samples
    )


def test_get_noise_samples(reparameterization):
    reparameterization_map = reparameterization.mapping
    samples = reparameterization.samples

    assert len(reparameterization_map) == len(samples)

    for noise, sample in zip(reparameterization_map, samples):
        assert noise[0] == sample[0]
        assert len(noise[1]) == len(sample[1])

        for (_, shape), xi_noise in zip(noise[1], sample[1]):
            assert isinstance(xi_noise, tf.Tensor)
            if shape == []:
                shape = xi_noise.shape.as_list()[-1:]
            assert list(xi_noise.shape) == [BATCH_SIZE, HORIZON] + shape


def test_encode_noise_samples_as_inputs(reparameterization):
    samples = reparameterization.samples
    encoding = reparameterization.encoding
    inputs = reparameterization.inputs

    samples_dict = dict(samples)
    encoding_dict = dict(encoding)

    for name, samples_lst in samples:
        if samples_lst:
            assert name in encoding_dict
        else:
            assert name not in encoding_dict

    i = 0
    for name, slices in encoding:
        samples_lst = samples_dict[name]
        assert len(slices) == len(samples_lst)

        for (start, end, slice_shape), sample in zip(slices, samples_lst):
            sample_shape = sample.shape.as_list()
            assert slice_shape == sample_shape[2:]
            assert start == i
            assert end == i + (end - start)
            assert end == i + np.prod(slice_shape) - 1

            i += end - start + 1

    assert isinstance(inputs, tf.Tensor)
    assert inputs.shape.as_list() == [BATCH_SIZE, HORIZON, i]


def test_decode_inputs_as_noise_samples(reparameterization):
    samples = reparameterization.samples
    decoded_samples = reparameterization.decoded_samples

    decoded_samples_dict = dict(decoded_samples)
    for name, samples_lst in samples:
        if samples_lst:
            assert name in decoded_samples_dict
        else:
            assert name not in decoded_samples_dict

    samples_dict = dict(samples)
    for name, decoded_sample_lst in decoded_samples_dict.items():
        sample_lst = samples_dict[name]
        assert len(decoded_sample_lst) == len(sample_lst)

        for decoded_sample, sample in zip(decoded_sample_lst, sample_lst):
            sample = sample[:, 0, ...]
            assert decoded_sample.dtype == sample.dtype
            assert decoded_sample.shape.as_list() == sample.shape.as_list()


def test_evaluate_noise_samples_as_inputs(reparameterization):
    compiler = reparameterization.compiler
    samples = reparameterization.samples
    inputs = reparameterization.inputs

    with tf.Session(graph=compiler.graph) as sess:
        noise_inputs = utils.evaluate_noise_samples_as_inputs(sess, samples)
        assert isinstance(noise_inputs, np.ndarray)
        assert list(noise_inputs.shape) == inputs.shape.as_list()



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.


"""Collection of reparameterization utility functions."""

# pylint: disable=missing-docstring


import numpy as np
import tensorflow as tf


def cell_size(sizes):
    return tuple(sz if sz != () else (1,) for sz in sizes)


def to_tensor(fluents):
    return tuple((f.tensor,) for f in fluents)


def get_noise_samples(reparameterization_map, batch_size, horizon):
    samples = []

    for name, noise_lst in reparameterization_map:
        cpf_scope = name.replace("'", "").replace("/", "_")

        sample_lst = []

        with tf.name_scope(cpf_scope):
            for dist, shape in noise_lst:
                sample_shape = [batch_size, horizon] + shape
                xi_noise = dist.sample(sample_shape)
                sample_lst.append(xi_noise)

        samples.append((name, sample_lst))

    return samples


def encode_noise_samples_as_inputs(samples):
    xi_placeholders = []
    encoding = []

    i = 0
    for name, xi_lst in samples:
        if not xi_lst:
            continue

        slices = []

        for xi_noise in xi_lst:
            batch_size = xi_noise.shape[0]
            horizon = xi_noise.shape[1]
            xi_shape = xi_noise.shape.as_list()[2:]
            xi_size = np.prod(xi_shape)

            xi_noise = tf.placeholder(
                xi_noise.dtype, shape=(batch_size, horizon, xi_size)
            )
            xi_placeholders.append(xi_noise)

            slices.append((i, i + xi_size - 1, xi_shape))
            i += xi_size

        encoding.append((name, slices))

    inputs = tf.concat(xi_placeholders, axis=2)

    return (inputs, encoding)


def decode_inputs_as_noise_samples(inputs, encoding):
    samples = []

    for name, slices in encoding:
        xi_lst = []

        for start, end, shape in slices:
            xi_noise = inputs[:, start : end + 1]
            xi_noise = tf.reshape(xi_noise, [-1, *shape])
            xi_lst.append(xi_noise)

        samples.append((name, xi_lst))

    return samples


def evaluate_noise_samples_as_inputs(sess, samples):
    xi_values = []

    for _, xi_lst in samples:
        if not xi_lst:
            continue

        for xi_noise in xi_lst:
            xi_noise = sess.run(xi_noise)
            xi_values.append(xi_noise)

    noise_inputs = np.concatenate(xi_values, axis=2)
    return noise_inputs



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-locals


"""tfplan CLI application."""

import json

import click
import psutil


@click.command()
@click.argument(
    "planner", type=click.Choice(["tensorplan", "straightline", "hindsight"])
)
@click.argument("rddl")
@click.option(
    "--batch-size",
    "-b",
    default=128,
    help="Number of trajectories in a batch.",
    show_default=True,
)
@click.option(
    "--horizon", "-hr", default=40, help="Number of timesteps.", show_default=True
)
@click.option(
    "--epochs", "-e", default=500, help="Number of training epochs.", show_default=True
)
@click.option(
    "--optimizer",
    type=click.Choice(
        [
            "Adadelta",
            "Adagrad",
            "Adam",
            "GradientDescent",
            "ProximalGradientDescent",
            "ProximalAdagrad",
            "RMSProp",
        ]
    ),
    default="GradientDescent",
    show_default=True,
)
@click.option(
    "--learning-rate",
    "-lr",
    default=1e-3,
    help="Optimizer's learning rate.",
    show_default=True,
)
@click.option(
    "--num-samples",
    "-n",
    type=int,
    default=1,
    help="Number of runs.",
    show_default=True,
)
@click.option(
    "--num-workers",
    type=click.IntRange(min=1, max=psutil.cpu_count()),
    default=1,
    help=f"Number of worker processes (min=1, max={psutil.cpu_count()}).",
    show_default=True,
)
@click.option(
    "--logdir",
    type=click.Path(),
    default="/tmp/tfplan/",
    help="Directory used for logging training summaries.",
    show_default=True,
)
@click.option("--config", type=click.File("r"), help="Configuration JSON file.")
@click.option("-v", "--verbose", is_flag=True, help="Verbosity flag.")
@click.version_option()
def cli(*args, **kwargs):
    """
    Planning via gradient-based optimization in TensorFlow.

    \b
    Args:
        RDDL Filename or rddlgym domain/instance id.
    """
    import functools
    import multiprocessing
    import os
    import re

    import numpy as np
    import pandas as pd
    from tqdm import tqdm

    config = kwargs

    if kwargs["config"]:
        json_config = json.load(kwargs["config"])
        config.update(json_config)
        del config["config"]

    config["optimization"] = {
        "optimizer": config["optimizer"],
        "learning_rate": config["learning_rate"],
    }

    n_samples = kwargs["num_samples"]
    num_workers = kwargs["num_workers"]

    previous_runs = [
        path for path in os.listdir(config["logdir"]) if re.search(r"run\d+$", path)
    ]
    start_id = len(previous_runs)
    pids = range(start_id, start_id + n_samples)

    pool = multiprocessing.Pool(
        processes=num_workers, initializer=tqdm.set_lock, initargs=(tqdm.get_lock(),)
    )
    trajectories = pool.map(functools.partial(run, config), pids)
    pool.close()
    pool.join()

    avg_uptime = np.mean([uptime for _, uptime, _ in trajectories])
    stddev_uptime = np.std([uptime for _, uptime, _ in trajectories])
    results = pd.concat([stats for _, _, stats in trajectories])
    results = results.groupby(results.index, sort=False).mean()
    print()
    print(f"===== Average ===== ({avg_uptime:.4f} ± {stddev_uptime:.4f} sec)")
    print(results)
    print()

    if kwargs["verbose"]:
        for i, (pid, uptime, stats) in enumerate(trajectories):
            print(f"===== Run #{i} / pid={pid} ({uptime:.4f} sec) =====")
            print(stats)
            print()


def run(config, n):
    import os
    import time

    # os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    os.environ["OMP_NUM_THREADS"] = str(psutil.cpu_count(logical=False))

    import rddlgym
    import tfplan

    import tensorflow as tf

    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

    config["run_id"] = n

    planner = config["planner"]
    rddl = config["rddl"]
    horizon = config["horizon"]
    debug = config["verbose"]

    config["logdir"] = os.path.join(config["logdir"], f"run{n}")
    filepath = os.path.join(config["logdir"], "data.csv")

    start = time.time()

    env = rddlgym.make(rddl, mode=rddlgym.GYM, config=config)
    env.set_horizon(horizon)

    planner = tfplan.make(planner, rddl, config)

    with rddlgym.Runner(env, planner, debug=debug) as runner:
        trajectory = runner.run()
        df = trajectory.save(filepath)
        stats = df.describe()

    uptime = time.time() - start

    pid = os.getpid()

    return pid, uptime, stats



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


import tensorflow as tf


class OpenLoopPolicy:
    """OpenLoopPolicy returns an action independently of the current state.

    Note:
        It uses the current state only for constraining the bounds of each action fluent.

    Args:
        compiler (:obj:`rddl2tf.compiler.Compiler`): A RDDL2TensorFlow compiler.
        batch_size (int): The simulation batch size.
        horizon(int): The number of timesteps.
    """

    def __init__(self, compiler, horizon, parallel_plans=True):
        self._compiler = compiler
        self.horizon = horizon
        self.parallel_plans = parallel_plans

        self._policy_variables = None

    @property
    def graph(self):
        """Returns the compiler's graph."""
        return self._compiler.graph

    @property
    def batch_size(self):
        """Returns the compiler's batch size."""
        return self._compiler.batch_size

    def build(self, scope, initializers=None):
        """Builds the policy."""
        with self.graph.as_default():
            with tf.variable_scope(scope):
                self._build_policy_variables(initializers)

    def _build_policy_variables(self, initializers=None):
        """Builds the policy variables for each action fluent."""
        action_fluents = self._compiler.rddl.domain.action_fluent_ordering
        action_size = self._compiler.rddl.action_size
        if initializers is None:
            initializers = [None] * len(action_fluents)
        self._policy_variables = []
        for fluent, shape, init in zip(action_fluents, action_size, initializers):
            var = self._get_policy_variable(fluent, shape, init)
            self._policy_variables.append(var)
        self._policy_variables = tuple(self._policy_variables)

    def __getitem__(self, i):
        return [var[i, ...] for var in self._policy_variables]

    def __call__(self, state, timestep):
        """Returns action fluents for the current `state` and `timestep`.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            timestep (tf.Tensor): The current timestep.

        Returns:
            Sequence[tf.Tensor]: A tuple of action fluents.
        """
        # pylint: disable=too-many-locals
        action_fluents = self._compiler.rddl.domain.action_fluent_ordering
        action_size = self._compiler.rddl.action_size
        bounds = self._compiler.action_bound_constraints(state)
        action = []
        with self.graph.as_default():
            timestep = tf.cast(timestep[0][0], tf.int32)
            for fluent, size, var in zip(
                action_fluents, action_size, self._policy_variables
            ):
                lower, upper = bounds[fluent]
                lower_batch = lower.batch if lower is not None else False
                upper_batch = upper.batch if upper is not None else False
                bounds_batch = lower_batch or upper_batch
                tensor = self._get_action_tensor(var[:, timestep, ...], (lower, upper))
                if not self.parallel_plans and not bounds_batch:
                    multiples = [self.batch_size] + [1] * len(size)
                    tensor = tf.tile(tensor, multiples)
                action.append(tensor)
        return tuple(action)

    def _get_policy_variable(self, fluent, fluent_shape, initializer=None):
        """Returns the correspondig policy variable for `fluent` with `fluent_shape`.

        Args:
            fluent (str): The fluent name.
            fluent_shape (Sequence[int]): The fluent shape.

        Returns:
            tf.Tensor: The policy variable for the action fluent.
        """
        if self.parallel_plans:
            shape = [self.batch_size, self.horizon] + list(fluent_shape)
        else:
            shape = [1, self.horizon] + list(fluent_shape)

        name = fluent.replace("/", "-")
        if initializer is not None:
            initializer = tf.constant_initializer(initializer, dtype=tf.float32)
        return tf.get_variable(
            name, dtype=tf.float32, shape=shape, initializer=initializer
        )

    @staticmethod
    def _get_action_tensor(policy_variable, bounds):
        """Returns the action tensor for `policy_variable` with domain
        constrainted by the action fluent precondition `bounds`

        Args:
            policy_variable (tf.Tensor): The policy variable.
            bounds (Tuple[Optional[rddl2t.core.fluent.TensorFluent],
            Optional[rddl2tf.core.fluent.TensorFluent]]): The (lower, upper) bounds.

        Returns:
            tf.Tensor: The action fluent tensor.
        """
        lower, upper = bounds
        lower = lower.cast(tf.float32).tensor if lower is not None else None
        upper = upper.cast(tf.float32).tensor if upper is not None else None
        if lower is not None and upper is not None:
            return lower + (upper - lower) * tf.sigmoid(policy_variable)
        if lower is not None and upper is None:
            return lower + tf.exp(policy_variable)
        if lower is None and upper is not None:
            return upper - tf.exp(policy_variable)
        return policy_variable



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


import os
from setuptools import setup, find_packages

import tfplan.version


def read(filename):
    filepath = os.path.join(os.path.dirname(__file__), filename)
    file = open(filepath, "r")
    return file.read()


setup(
    name="tf-plan",
    version=tfplan.version.__version__,
    author="Thiago P. Bueno",
    author_email="thiago.pbueno@gmail.com",
    description="Planning through backpropagation using TensorFlow.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="GNU General Public License v3.0",
    keywords=["planning", "tensorflow", "rddl", "mdp"],
    url="https://github.com/thiagopbueno/tf-plan",
    packages=find_packages(),
    entry_points="""
        [console_scripts]
        tfplan=scripts.tfplan:cli
    """,
    python_requires=">=3.5",
    install_requires=[
        "Click",
        "numpy",
        "tqdm",
        "psutil",
        "tensorflow-cpu==1.15",
        "rddlgym",
        "rddl2tf",
    ],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring,redefined-outer-name


from collections import OrderedDict
import pytest
import numpy as np
import tensorflow as tf

import rddlgym
from tfplan.planners import DEFAULT_CONFIG, HindsightPlanner
from tfplan.planners.stochastic import utils


BATCH_SIZE = 16
HORIZON = 20
EPOCHS = 5


@pytest.fixture(scope="module", params=["Navigation-v2"])
def planner(request):
    rddl = request.param
    config = {
        **DEFAULT_CONFIG,
        **{"batch_size": BATCH_SIZE, "horizon": HORIZON, "epochs": EPOCHS},
        "verbose": False,
    }
    planner = HindsightPlanner(rddl, config)
    planner.build()
    return planner


def test_build_base_policy_ops(planner):
    base_policy = planner.base_policy
    assert not base_policy.parallel_plans
    assert base_policy.horizon == 1
    assert hasattr(base_policy, "_policy_variables")


def test_build_scenario_policy_ops(planner):
    scenario_policy = planner.scenario_policy
    compiler = planner.compiler
    assert scenario_policy.parallel_plans
    assert scenario_policy.horizon == compiler.rddl.instance.horizon - 1
    assert hasattr(scenario_policy, "_policy_variables")


def test_build_initial_state_ops(planner):
    initial_state = planner.initial_state
    compiler = planner.compiler
    batch_size = compiler.batch_size
    assert isinstance(initial_state, tuple)
    assert len(initial_state) == len(compiler.initial_state_fluents)
    for tensor, fluent in zip(initial_state, compiler.initial_state_fluents):
        assert tensor.dtype == fluent[1].dtype
        assert tensor.shape == (batch_size, *fluent[1].shape.fluent_shape)


def test_build_scenario_start_states_ops(planner):
    initial_state = planner.initial_state
    next_state = planner.next_state
    assert isinstance(next_state, tuple)
    assert len(next_state) == len(initial_state)
    for initial_state_tensor, next_state_tensor in zip(initial_state, next_state):
        assert initial_state_tensor.shape == next_state_tensor.shape
        assert initial_state_tensor.dtype == next_state_tensor.dtype


def test_build_sequence_length_ops(planner):
    assert planner.steps_to_go is not None
    assert planner.steps_to_go.dtype == tf.int32
    assert planner.steps_to_go.shape == ()
    assert planner.sequence_length is not None
    assert planner.sequence_length.dtype == tf.int32
    assert planner.sequence_length.shape == (planner.compiler.batch_size,)


def test_build_trajectory_ops(planner):
    trajectory = planner.trajectory
    actions = trajectory.actions
    rewards = trajectory.rewards
    assert rewards.shape == (BATCH_SIZE, HORIZON - 1, 1)

    action_fluents = planner.compiler.default_action_fluents
    for action, action_fluent in zip(actions, action_fluents):
        size = action_fluent[1].shape.as_list()
        assert action.shape.as_list() == [BATCH_SIZE, HORIZON - 1, *size]

    with planner.graph.as_default():
        last_reward = tf.reduce_mean(planner.trajectory.rewards[:, -1, 0])
        base_policy_vars = tf.trainable_variables(scope="base_policy")
        base_policy_grads = tf.gradients(last_reward, base_policy_vars)

        base_policy_grads_ = _session_run(planner, base_policy_grads)
        for grad_ in base_policy_grads_:
            assert grad_ is not None
            assert not np.allclose(grad_, np.zeros_like(grad_))

        reward = tf.reduce_mean(planner.reward)
        base_policy_grads = tf.gradients(reward, base_policy_vars)
        assert all(grad is None for grad in base_policy_grads)


def test_loss_ops(planner):
    reward = planner.reward
    scenario_total_reward = planner.scenario_total_reward

    assert reward.shape == (BATCH_SIZE,)
    assert scenario_total_reward.shape == (BATCH_SIZE,)


def test_optimization_ops(planner):

    with planner.graph.as_default():

        grads_and_vars = planner.grads_and_vars
        assert isinstance(grads_and_vars, list)
        assert len(grads_and_vars) == len(tf.trainable_variables())

        for variable in tf.trainable_variables(scope="base_policy"):
            assert variable.shape[:2] == [1, 1]

        for variable in tf.trainable_variables(scope="scenario_policy"):
            assert variable.shape[:2] == [BATCH_SIZE, HORIZON - 1]

    grads_and_vars_ = _session_run(planner, grads_and_vars)
    for grad_, _ in grads_and_vars_:
        assert grad_ is not None
        assert not np.allclose(grad_, np.zeros_like(grad_))


def test_call(planner):
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)
    state, timestep = env.reset()
    action = planner(state, timestep)
    assert isinstance(action, OrderedDict)


def test_get_batch_initial_state(planner):
    # pylint: disable=protected-access
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)

    with planner.compiler.graph.as_default():
        state = env.observation_space.sample()
        batch_state = planner._get_batch_initial_state(state)
        assert len(state) == len(batch_state)

        for fluent, batch_fluent in zip(state.values(), batch_state):
            assert fluent.dtype == batch_fluent.dtype
            assert fluent.shape == batch_fluent.shape[1:]
            assert batch_fluent.shape[0] == planner.compiler.batch_size


def test_get_action(planner):
    # pylint: disable=protected-access
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)

    with tf.Session(graph=planner.compiler.graph) as sess:
        sess.run(tf.global_variables_initializer())
        feed_dict = _get_feed_dict(sess, planner, env)

        actions_ = planner._get_action(planner.action, feed_dict)

        action_fluents = planner.compiler.default_action_fluents
        assert isinstance(actions_, OrderedDict)
        assert len(actions_) == len(action_fluents)
        for action_, action_fluent in zip(actions_.values(), action_fluents):
            assert tf.dtypes.as_dtype(action_.dtype) == action_fluent[1].dtype
            assert list(action_.shape) == list(action_fluent[1].shape.fluent_shape)


def test_runner(planner):
    # pylint: disable=protected-access
    rddl = planner.rddl
    env = rddlgym.make(rddl, mode=rddlgym.GYM)
    env._horizon = 3
    runner = rddlgym.Runner(env, planner)
    trajectory = runner.run()
    assert len(trajectory) == env._horizon


def _session_run(planner, fetches):
    env = rddlgym.make(planner.rddl, mode=rddlgym.GYM)

    with tf.Session(graph=planner.compiler.graph) as sess:
        sess.run(tf.global_variables_initializer())
        feed_dict = _get_feed_dict(sess, planner, env)
        return sess.run(fetches, feed_dict=feed_dict)


def _get_feed_dict(sess, planner, env):
    # pylint: disable=protected-access
    state = env.observation_space.sample()
    batch_state = planner._get_batch_initial_state(state)

    next_state_noise = utils.evaluate_noise_samples_as_inputs(
        sess, planner.cell_samples
    )
    scenario_noise = utils.evaluate_noise_samples_as_inputs(
        sess, planner.simulator.samples
    )

    feed_dict = {
        planner.initial_state: batch_state,
        planner.cell_noise: next_state_noise,
        planner.simulator.noise: scenario_noise,
        planner.steps_to_go: HORIZON - 1,
    }
    return feed_dict



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring

import tensorflow as tf

from rddl2tf.compilers import ReparameterizationCompiler

from tfplan.planners.stochastic import StochasticPlanner
from tfplan.train.policy import OpenLoopPolicy
from tfplan.planners.stochastic.simulation import Simulator
from tfplan.planners.stochastic import utils


class StraightLinePlanner(StochasticPlanner):
    """StraightLinePlanner class implements the online gradient-based
    planner that chooses the next action based on the lower bound of
    the Value function of the start state.

    Args:
        rddl (str): A RDDL domain/instance filepath or rddlgym id.
        config (Dict[str, Any]): The planner config dict.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, rddl, config):
        super().__init__(rddl, ReparameterizationCompiler, config)

        self.policy = None

        self.simulator = None
        self.trajectory = None
        self.final_state = None
        self.total_reward = None

        self.avg_total_reward = None
        self.loss = None

        self.writer = None
        self.summaries = None

    def _build_policy_ops(self):
        horizon = self.config["horizon"]
        self.policy = OpenLoopPolicy(self.compiler, horizon, parallel_plans=False)
        self.policy.build("planning")

    def _build_trajectory_ops(self):
        with tf.name_scope("scenarios"):
            self.simulator = Simulator(self.compiler, self.policy, config=None)
            self.simulator.build()
            (
                self.trajectory,
                self.final_state,
                self.total_reward,
            ) = self.simulator.trajectory(self.initial_state, self.sequence_length)

    def _build_loss_ops(self):
        with tf.name_scope("loss"):
            self.avg_total_reward = tf.reduce_mean(self.total_reward)
            self.loss = tf.square(self.avg_total_reward)

    def _build_summary_ops(self):
        if self.config["verbose"]:

            with tf.name_scope("summary"):
                _ = tf.compat.v1.summary.FileWriter(self.config["logdir"], self.graph)
                tf.compat.v1.summary.scalar("avg_total_reward", self.avg_total_reward)
                tf.compat.v1.summary.scalar("loss", self.loss)

                tf.compat.v1.summary.histogram("total_reward", self.total_reward)
                tf.compat.v1.summary.histogram("scenario_noise", self.simulator.noise)

                for grad, variable in self.grads_and_vars:
                    var_name = variable.name
                    tf.compat.v1.summary.histogram(f"{var_name}_grad", grad)
                    tf.compat.v1.summary.histogram(var_name, variable)

                self.summaries = tf.compat.v1.summary.merge_all()

    def __call__(self, state, timestep):
        scenario_noise = utils.evaluate_noise_samples_as_inputs(
            self._sess, self.simulator.samples
        )

        feed_dict = {
            self.initial_state: self._get_batch_initial_state(state),
            self.simulator.noise: scenario_noise,
            self.steps_to_go: self.config["horizon"] - timestep,
        }

        self.run(timestep, feed_dict)

        action = self._get_action(self.trajectory.actions, feed_dict)
        return action



# This file is part of tf-plan.

# tf-plan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-plan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-plan. If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring


def make(planner, rddl, config):
    """tf-plan planner factory."""
    from tfplan.planners import Tensorplan, StraightLinePlanner, HindsightPlanner

    planner_cls = {
        "tensorplan": Tensorplan,
        "straightline": StraightLinePlanner,
        "hindsight": HindsightPlanner,
    }

    return planner_cls[planner](rddl, config)
